/*
 *      b32api.c
 *
 *      Copyright (c) 1994 by Borland International
 *      All Rights Reserved.
 *
 */

#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include <dir.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>

#include "basetype.h"
#include "oldexe.h"
#include "pehdr.h"

static bits8				*fullHeader;
static PEObjectTabEntry		*pImp;

static void usage (void)
{
	printf ("\nB32API <filespec>[.DLL]\n"
			"  Prints compatibility report of 32bit EXE or DLL\n");
}

static void setExtension (char *dest, const char *path, const char *newExt, int force)
{
    char	drive	[MAXDRIVE];
    char	direct	[MAXDIR];
    char	name	[MAXFILE];
    char	ext		[MAXEXT];

	fnsplit (path, drive, direct, name, ext);
	if	(force || !*ext)
		fnmerge (dest, drive, direct, name, newExt);
	else
		fnmerge (dest, drive, direct, name, ext);
}

static bits8 *loadIData (FILE *fp, PEHeader *hdr)
{
	PEObjectTabEntry	*pObj;
	bits8				*idata;
	int					i;

	fullHeader = (bits8 *)malloc (hdr->headerSize);
	assert (fullHeader);
	fread (fullHeader, hdr->headerSize, 1, fp);

	pObj = (PEObjectTabEntry *)fullHeader;
	for (i = 0; i < hdr->nObjects; i++)
	{
		if	(strcmpi ((char *)pObj->name, ".idata") == 0)
			break;
		pObj++;
	}

	/*
	/*  There better have been a .idata section.
	*/
	assert (i <= hdr->nObjects);

	pImp = pObj;

	idata = (bits8 *)malloc (pObj->physicalSize);
	fseek (fp, pImp->physicalOffset, SEEK_SET);
	fread (idata, pObj->physicalSize, 1, fp);

	return idata;
}

static boolean compareModules (char *name, char **modules)
{
	while	(*modules)
	{
		if	(!stricmp (name, *modules++))
			return TRUE;
	}
	return FALSE;
}

static boolean systemModule (char *name)
{
	static char *systemModules [] =
	{
		"KERNEL32.dll",
		"NTDLL.dll",
		"USER32.dll",
		"GDI32.dll",
		0
	};

	return compareModules (name, systemModules);
}

static boolean bigfootModule (char *name)
{
	static char *bigfootModules [] =
	{
		"KERNEL32.dll",
		"NTDLL.dll",
		"USER32.dll",
		0
	};

	return compareModules (name, bigfootModules);
}
			
static void reportAPIFlags (char *apiName, HMODULE modHandle)
{
	FARPROC	proc;

	proc = GetProcAddress (modHandle, apiName);
	if	(!proc)
		printf ("Not implemented\n");
	else
		printf ("Implemented\n");
}

static void reportImports (bits8 *buff)
{
	ImportDirectory	*pIDir;
	bits32			*impOrd;
	bits8			*hintName;
	HMODULE			modHandle;
	
	pIDir = (ImportDirectory *)buff;
	while (pIDir->nameRVA != 0)
	{
		char	*modName;

		modName = (char *)&buff[pIDir->nameRVA - pImp->RVA];

		/*
		/*  If this is not a Win32 System module, we just continue.
		/*
		*/
		if	(!systemModule (modName))
		{
			printf ("\nModule %s is not a system module.\n", modName);
			pIDir++;
			continue;
		}

		if	(!bigfootModule (modName))
		{
			printf ("\nThe Borland 32bit DOS extender does not implement %s\n", modName);
			pIDir++;
			continue;
		}

		printf ("\nImports from %s\n", modName);
		
		if	(pIDir->basic.flags)
			impOrd = (bits32 *)&buff[pIDir->basic.flags - pImp->RVA];
		else
			impOrd = (bits32 *)&buff[pIDir->lookupRVA - pImp->RVA];

		modHandle = GetModuleHandle (modName);
		assert (modHandle);
		
		while (*impOrd)
		{
			/*
			/*  Better not be any imports by ordinal, since MS screwed that up.
			*/
			assert (!(*impOrd & 0x80000000));
			
			hintName = buff + (*impOrd - pImp->RVA);
			printf ("%-40s", hintName + 2);
			reportAPIFlags ((char *)(hintName + 2), modHandle);
			impOrd++;
		}
		pIDir++;
	}
}

static int report (FILE *fp)
{
	OldExeHdr	oldHdr;
	PEHeader	hdr;
	bits8		*idata;
	
	fread (&oldHdr, sizeof (oldHdr), 1, fp);
	if	(oldHdr.signature != MZSIGNATURE	||
		 oldHdr.reloc_offset != 0x40)
	{
		printf ("Not a PE file\n");
		return 1;
	}

	fseek (fp, oldHdr.NE_LE_start, SEEK_SET);
	fread (&hdr, sizeof (hdr), 1, fp);
	if	(*(bits32 *)&hdr.signature [0] != PESIGNATURE)
	{
		printf ("Not a PE file\n");
		return 1;
	}

	idata = loadIData (fp, &hdr);

	reportImports (idata);

	free (idata);
	free (fullHeader);
	
	return 0;
}

static void banner (void)
{
	printf ("Borland 32bit API compatibility tester\n"
			"Version .01  Copyright (c) 1994 Borland International\n");
}

int main (int argc, char **argv)
{
	char		path [MAXPATH];
	FILE		*fp;
	int			res;
	HMODULE		kernelModule;

	banner ();
	
	if	(argc < 2)
	{
		usage ();
		return 0;
	}

	kernelModule = GetModuleHandle ("KERNEL32.DLL");
	assert (kernelModule);

	if	(!GetProcAddress (kernelModule, "Borland32"))
	{
		printf ("This application must be run under the Borland 32bit DOS extender.\n");
		return 2;
	}

	setExtension (path, argv [1], ".dll", 0);
	fp = fopen (path, "rb");
	if	(!fp)
	{
		printf ("Couldn't open '%s'\n", path);
		return 2;
	}

	res = report (fp);
	fclose (fp);

	return res;
}

/*
**  End.
*/

